from unittest import TestCase
import pytest
from bqplot import pyplot as plt
from BNumMet.Interpolation import polinomial, piecewise_linear, pchip, splines
from BNumMet.Visualizers.InterpolationVisualizer import InterpolVisualizer
import numpy as np


class test_PolyInterpolation(TestCase):
    def test_interpolation(self):
        """
        Test the interpolation of a polynomial of degree 3
        """
        x = np.array([1, 2, 3, 4])
        y = np.array([1, 4, 9, 16])
        u = np.array([1.5, 2.5, 3.5])
        v = polinomial(x, y, u)
        self.assertTrue(np.allclose(v, np.array([2.25, 6.25, 12.25])))
        self.assertTrue(np.allclose(polinomial(x, y, x), y))

    def test_notOrdered(self):
        """
        Test the interpolation of a polynomial of degree 3 with unsorted x - this should work aka True
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [6.25])
        u = np.arange(1, 4.1, 0.1)
        v = polinomial(x, y, u)

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertTrue(np.isclose(v[i1], 6.25))


class test_piecewise_linearInterpolation(TestCase):
    def test_interpolation(self):
        """
        Test the piecewise linear interpolation
        """
        x = np.array([1, 2, 3, 4])
        y = np.array([1, 4, 9, 16])
        u = np.arange(1, 4.2, 0.1)
        v = piecewise_linear(x, y, u)

        for i in range(len(x)):
            index = np.where(np.isclose(u, x[i]))[0][0]
            print(index, v[index], x[i], y[i])
            self.assertTrue(np.isclose(v[index], y[i]))

    def test_notOrdered(self):
        """
        Test the piecewise linear interpolation with unsorted x - this should NOT work aka is False
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = piecewise_linear(
            x, y, u, sorted=True
        )  # Sorted = True makes the assumption that x is sorted - IT IS NOT

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertFalse(np.isclose(v[i1], -5))

    def test_notOrdered_unsorted(self):
        """
        Test the piecewise linear interpolation with unsorted x - this should work aka True
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = piecewise_linear(x, y, u, sorted=False)

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertTrue(np.isclose(v[i1], -5))


class test_pchip(TestCase):
    def test_interpolation(self):
        """
        Test the pchip interpolation
        """
        x = np.array([1, 2, 3, 4])
        y = np.array([1, 4, 9, 16])
        u = np.arange(1, 4.2, 0.1)
        v = pchip(x, y, u)

        for i in range(len(x)):
            index = np.where(np.isclose(u, x[i]))[0][0]
            print(index, v[index], x[i], y[i])
            self.assertTrue(np.isclose(v[index], y[i]))

    def test_notOrdered(self):
        """
        Test the pchip interpolation with unsorted x - this should NOT work aka is False
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = pchip(
            x, y, u, sorted=True
        )  # Sorted = True makes the assumption that x is sorted - IT IS NOT

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertFalse(np.isclose(v[i1], -5))

    def test_notOrdered_unsorted(self):
        """
        Test the pchip interpolation with unsorted x - this should work aka True
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = pchip(x, y, u, sorted=False)

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertTrue(np.isclose(v[i1], -5))


class test_splines(TestCase):
    def test_interpolation(self):
        """
        Test the splines interpolation
        """
        x = np.array([1, 2, 3, 4])
        y = np.array([1, 4, 9, 16])
        u = np.arange(1, 4.2, 0.1)
        v = splines(x, y, u)

        for i in range(len(x)):
            index = np.where(np.isclose(u, x[i]))[0][0]
            print(index, v[index], x[i], y[i])
            self.assertTrue(np.isclose(v[index], y[i]))

    def test_notOrdered(self):
        """
        Test the splines interpolation with unsorted x - this should NOT work aka is False
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = splines(
            x, y, u, sorted=True
        )  # Sorted = True makes the assumption that x is sorted - IT IS NOT

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertFalse(np.isclose(v[i1], -5))

    def test_notOrdered_unsorted(self):
        """
        Test the splines interpolation with unsorted x - this should work aka True
        """
        x = np.array([1, 2, 3, 4] + [2.5])
        y = np.array([1, 4, 9, 16] + [-5])
        u = np.arange(1, 4.1, 0.1)
        v = splines(x, y, u, sorted=False)

        # find the index of the point 2.5 on u
        i1 = np.where(np.isclose(u, 2.5))[0][0]

        print(i1, v[i1])

        self.assertTrue(np.isclose(v[i1], -5))


class test_InterpolationVisualizer(TestCase):
    # Run before each test

    def runtest_setup(self):
        try:
            plt.close(self.interpolVisualizer.fig)
        except AttributeError:
            pass

        x = list(np.arange(1, 7, 1).astype(float))
        y = np.array([16, 18, 21, 17, 15, 12], dtype=float)
        u = list(np.arange(1, 6.1, 0.1).astype(float))

        self.interpolVisualizer = InterpolVisualizer(x, y, u)
        self.interpolVisualizer.run()

        print("Setting Up - Interpolation Visualizer Test")

    def test_no_param_init(self):
        """
        Test the interpolation visualizer with no parameters passed in - should use default values for x, y, u
        """
        x = list(np.arange(1, 7, 1).astype(float))
        y = np.array([16, 18, 21, 17, 15, 12], dtype=float)
        u = list(np.arange(1, 6.1, 0.1).astype(float))

        interpolVisualizer = InterpolVisualizer()
        interpolVisualizer.run()

        self.assertTrue(all(interpolVisualizer.x == x))
        self.assertTrue(all(interpolVisualizer.y == y))
        self.assertTrue(all(interpolVisualizer.u == u))

    def test_raiseException_lenXnotEqualLenY(self):
        self.runtest_setup()

        x = [1, 2, 3, 4]
        y = [1, 4, 9, 16, 25]
        u = [1, 2, 3, 4, 5]

        with self.assertRaises(ValueError):
            InterpolVisualizer(x, y, u)

    def test_updatePoints(self):
        self.runtest_setup()

        oldX = self.interpolVisualizer.scattered_dots.x
        oldY = self.interpolVisualizer.scattered_dots.y

        # 1. Adding a point will correctly update the scattered dots
        self.interpolVisualizer.scattered_dots.x = np.append(
            self.interpolVisualizer.scattered_dots.x, 7
        )
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, 7
        )
        self.assertTrue(
            len(self.interpolVisualizer.scattered_dots.x) == len(oldX) + 1
            and len(self.interpolVisualizer.scattered_dots.y) == len(oldY) + 1
        )

        # 2. Adding same Val of X will not update the scattered dots
        self.interpolVisualizer.scattered_dots.x = np.append(
            self.interpolVisualizer.scattered_dots.x, 7
        )
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, 7
        )
        self.assertTrue(
            len(self.interpolVisualizer.scattered_dots.x) == len(oldX) + 1
            and len(self.interpolVisualizer.scattered_dots.y) == len(oldY) + 1
        )

        # 3. Adding Y without X will not update the scattered dots
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, 7
        )
        self.assertTrue(len(self.interpolVisualizer.scattered_dots.y) == len(oldY) + 1)

    def test_block_adding(self):
        self.runtest_setup()

        oldX = self.interpolVisualizer.scattered_dots.x
        oldY = self.interpolVisualizer.scattered_dots.y

        # 1. Adding a point will correctly update the scattered dots
        self.interpolVisualizer.scattered_dots.x = np.append(
            self.interpolVisualizer.scattered_dots.x, 7
        )
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, 7
        )
        self.assertTrue(
            len(self.interpolVisualizer.scattered_dots.x) == len(oldX) + 1
            and len(self.interpolVisualizer.scattered_dots.y) == len(oldY) + 1
        )

        # 2. Adding some points with "block_adding" - checkbox - will not update the scattered dots
        self.interpolVisualizer.block_adding.value = True
        self.interpolVisualizer.scattered_dots.x = np.append(
            self.interpolVisualizer.scattered_dots.x, [8, 9, 10]
        )
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, [8, 9, 10]
        )
        self.assertTrue(
            len(self.interpolVisualizer.scattered_dots.x) == len(oldX) + 1
            and len(self.interpolVisualizer.scattered_dots.y) == len(oldY) + 1
        )

    def test_extrapolationEffects(self):
        self.runtest_setup()

        # 1. Extrapolation checkbox is NOT checked by default
        self.assertFalse(self.interpolVisualizer.extrapolation.value)

        # 2. Extrapolation checkbox is checked - u is updated (different min and max)
        minMaxX = [min(self.interpolVisualizer.x), max(self.interpolVisualizer.x)]
        self.interpolVisualizer.extrapolation.value = True
        self.assertTrue(self.interpolVisualizer.extrapolation.value)
        self.assertTrue(
            self.interpolVisualizer.u[0] < minMaxX[0]
            and self.interpolVisualizer.u[-1] > minMaxX[1]
        )

        # 3. Extrapolation checkbox is NOT checked - u is updated (same min and max)
        self.interpolVisualizer.extrapolation.value = False
        self.assertFalse(self.interpolVisualizer.extrapolation.value)
        print(min(self.interpolVisualizer.u), minMaxX[0])
        print(max(self.interpolVisualizer.u), minMaxX[1])
        self.assertTrue(
            min(self.interpolVisualizer.u) == minMaxX[0]
            and max(self.interpolVisualizer.u) == minMaxX[1]
        )

    def test_updateCheckboxes(self):
        self.runtest_setup()

        self.interpolVisualizer.interpol_lines()
        self.assertTrue(
            len(self.interpolVisualizer.interpolation_lines)
            == len(self.interpolVisualizer.methods)
        )

        self.interpolVisualizer.checkboxes[0].value = False
        self.assertTrue(
            len(self.interpolVisualizer.interpolation_lines)
            == len(self.interpolVisualizer.methods) - 1
        )

    def test_reset(self):
        self.runtest_setup()

        oldX = self.interpolVisualizer.scattered_dots.x
        oldY = self.interpolVisualizer.scattered_dots.y

        # 1. Adding a point will correctly update the scattered dots
        self.interpolVisualizer.scattered_dots.x = np.append(
            self.interpolVisualizer.scattered_dots.x, 7
        )
        self.interpolVisualizer.scattered_dots.y = np.append(
            self.interpolVisualizer.scattered_dots.y, 7
        )

        self.interpolVisualizer.reset(None)

        self.assertTrue(
            (self.interpolVisualizer.scattered_dots.x == oldX).all()
            and (self.interpolVisualizer.scattered_dots.y == oldY).all()
        )

    def test_autoZoom(self):
        self.runtest_setup()

        # 1. AutoZoom button is clicked without Effects of Extrapolation - sc.x and sc.y limits are updated
        self.interpolVisualizer.auto_zoom_button.click()
        self.assertTrue(
            self.interpolVisualizer.x_sc.min == min(self.interpolVisualizer.u)
            and self.interpolVisualizer.x_sc.max == max(self.interpolVisualizer.u)
        )

        # 2. AutoZoom button is clicked with Effects of Extrapolation - sc.x and sc.y limits are updated
        self.interpolVisualizer.extrapolation.value = True
        self.interpolVisualizer.auto_zoom_button.click()
        self.assertTrue(
            self.interpolVisualizer.x_sc.min == min(self.interpolVisualizer.u)
            and self.interpolVisualizer.x_sc.max == max(self.interpolVisualizer.u)
        )



from unittest import TestCase
import pytest


class General(TestCase):
    def test_import(self):
        import BNumMet

        assert BNumMet

    def test_import_visualizers(self):
        import BNumMet.Visualizers

        assert BNumMet.Visualizers

    def test_import_linearsystems_all(self):
        from BNumMet import LinearSystems

        assert LinearSystems.lu
        assert LinearSystems.interactive_lu
        assert LinearSystems.permute

    def test_import_interpolation_all(self):
        from BNumMet import Interpolation

        assert Interpolation.polinomial
        assert Interpolation.piecewise_linear
        assert Interpolation.pchip
        assert Interpolation.splines

    def test_import_nonlinear_all(self):
        from BNumMet import NonLinear

        assert NonLinear.bisect
        assert NonLinear.secant
        assert NonLinear.newton
        assert NonLinear.IQI
        assert NonLinear.zBrentDekker



import numpy as np
from BNumMet.LinearSystems import lu_solve


def polinomial(x, y, u):
    """
    Computes the polynomial interpolation of a set of points (x,y) at the points u

    params:
        x: list of x coordinates  # list of x coordinates of the input data points
        y: list of y coordinates  # list of y coordinates of the input data points
        u: list of points where the interpolation is computed  # list of points where the interpolation is to be computed

    returns:
        v: list of values of the interpolation at the points u  # list of interpolated values at the specified points u
    """
    n = len(x)  # number of input data points
    v = np.zeros(
        len(u)
    )  # Create a vector of zeros of the same size as u, to store the result
    for i in range(n):
        w = np.ones(len(u))  # initializing the weight vector with ones
        for j in range(n):
            if j != i:  # exclude current data point from calculation of weights
                w *= (u - x[j]) / (
                    x[i] - x[j]
                )  # calculate weight for current data point
        v += (
            w * y[i]
        )  # accumulate weighted y values to get the final interpolated value

    return v  # return the final interpolated values


def piecewise_linear(x, y, u, sorted=False):
    """
    Computes the piecewise lineal interpolation of a set of points (x,y) at the points u, (x,y) ACCORDING TO THE ALGORITHM IN THE BOOK

    params:
        x: list of x coordinates  # list of x coordinates of the input data points
        y: list of y coordinates  # list of y coordinates of the input data points
        u: list of points where the interpolation is computed  # list of points where the interpolation is to be computed
        sorted (optional): if the points are sorted or not (default: False)  # boolean flag indicating if the input data points are sorted or not

    returns:
        v: list of values of the interpolation at the points u  # list of interpolated values at the specified points u
    """
    if not sorted:  # if the input data points are not sorted
        # Sort the points
        x = np.array(x)  # convert x to numpy array
        y = np.array(y)  # convert y to numpy array
        ind = np.argsort(x)  # Get the indices of the sorted array
        x = x[ind]  # Sort the x coordinates
        y = y[ind]  # Sort the y coordinates

    delta = np.diff(y) / np.diff(
        x
    )  # Compute the slopes of the lines -- here we are using the fact that x is sorted
    n = len(x)  # number of input data points

    k = np.zeros(np.size(u), dtype=int)  # initialize array to store indices
    for j in np.arange(1, n - 1):
        k[
            x[j] <= u
        ] = j  # find the indices of the data points just smaller than the corresponding u values

    s = u - x[k]  # calculate the difference between u and corresponding x values
    v = (
        y[k] + s * delta[k]
    )  # calculate the interpolated y values based on the slope of the line

    return v  # return the final interpolated values


def pchip(x, y, u, sorted=False):
    """
    Piecewise Cubic Hermite Interpolation Polynomial (P.C.H.I.P.) [Based on an old Fortran program by Fritsch and Carlson]

        params:
            x: list of x coordinates
            y: list of y coordinates
            u: list of points where the interpolation is computed
            sorted (optional): if the points are sorted or not (default: False)

        returns:
            v: list of values of the interpolation at the points u
    """

    def pchip_end(h1, h2, delta1, delta2):
        """
        Computes the slopes at the end points of the interval

        params:
            x: list of x coordinates # x-coordinates of the data points
            y: list of y coordinates # y-coordinates of the data points
            u: list of points where the interpolation is computed # Points where the interpolation is to be computed
            sorted (optional): if the points are sorted or not (default: False) # Parameter to specify if the points are already sorted

        returns:
            v: list of values of the interpolation at the points u # List of values of the P.C.H.I.P. at the points in 'u'
        """
        # Noncenter, shape-preserving, three-point formula.
        d = ((2 * h1 + h2) * delta1 - h1 * delta2) / (h1 + h2)
        # If slopes of the secant lines are of different sign or If the slopes are not of the same magnitude, use 0.
        if (
            np.sign(delta1) != np.sign(delta2)
            or np.abs(d) > np.abs(3 * delta1)
            or np.abs(d) > np.abs(3 * delta2)
        ):
            d = 0

        return d

    def pchip_slopes(h, delta):
        """
        Slopes for shape-preserving Hermite cubic, computes the slopes
            - Interior Points
                * d(k) = 0 <- delta(k-1) && delta(k) different signs or both are 0
                * d(k) = Weighted Harmonic Mean <- Same sign delta(k-1) && delta (k)
            - EndPoints
                Call pchip end :)

        params:
            h: list of distances between points
            delta: list of slopes between points

        returns:
            d: list of slopes for the Hermite cubic
        """
        d = np.zeros(len(h))  # Initialize an array of zeros to store the slopes

        k = np.where(np.sign(delta[0:-1]) * np.sign(delta[1:]) > 0)[
            0
        ]  # Find the indices of the points where the slopes are of the same sign. 'k' will be an array of indices.
        k = k + 1  # Add 1 to the indices to get the indices of the slopes

        w1 = 2 * h[k] + h[k - 1]
        w2 = h[k] + 2 * h[k - 1]
        d[k] = (w1 + w2) / (
            w1 / delta[k - 1] + w2 / delta[k]
        )  # Compute the slopes of the lines for the interior points, where the slopes are of the same sign

        # end points
        d[0] = pchip_end(
            h[0], h[1], delta[0], delta[1]
        )  # Compute the slope of the first endpoint using the 'pchip_end' function

        d = np.append(
            d, pchip_end(h[-2], h[-3], delta[-2], delta[-3])
        )  # Compute the slope of the last endpoint using the 'pchip_end' function and append it to 'd'

        return d  # Return the list of slopes

    # Sort the points
    x = np.array(x)  # Convert the x coordinate input to a numpy array
    y = np.array(y)  # Convert the y coordinate input to a numpy array
    if not sorted:  # If the points are not already sorted
        ind = np.argsort(x)  # Get the indices of the sorted array
        x = x[ind]  # Sort the x coordinates
        y = y[ind]  # Sort the y coordinates

    # First derivative
    h = np.diff(x)  # Compute the distances between the points in x
    delta = np.diff(y) / h  # Compute the slopes between the points

    d = pchip_slopes(
        h, delta
    )  # Compute the slopes for the Hermite cubic using the pchip_slopes function

    n = len(x)
    c = (3 * delta - 2 * d[:-1] - d[1:]) / (
        h
    )  # Compute the coefficients of the cubic polynomials
    b = (d[:-1] - 2 * delta + d[1:]) / (
        h**2
    )  # Compute the coefficients of the cubic polynomials

    k = np.zeros(np.size(u), dtype=int)  # Initialize an array of indices
    for j in np.arange(1, n - 1):  # Loop through the indices
        k[x[j] <= u] = j  # Update the indices where x[j] is less than or equal to u

    s = u - x[k]  # Compute the value of s for each index
    v = y[k] + s * (
        d[k] + s * (c[k] + s * b[k])
    )  # Compute the values of the Hermite cubic

    return v  # Return the computed values


def splines(x, y, u, sorted=False):
    """
    Finds the piecewise cubic interpolatory spline S(x), with S(x(j)) = y(j), and returns v(k) = S(u(k)).

    params:
        x: list of x coordinates - list of x values to be used as input to the spline
        y: list of y coordinates - list of y values to be used as input to the spline
        u: list of points where the interpolation is computed - the list of x-coordinates where the spline should be evaluated
        sorted (optional): if the points are sorted or not (default: False) - flag to indicate whether the input points are sorted or not

    returns:
        v: list of values of the interpolation at the points u - the y-values of the spline evaluated at the x-coordinates in u
    """

    def splineslopes(h, delta):
        """
        Computes the slopes of the splines Uses not-a-knot end conditions.

        params:
            h: list of distances between points
            delta: list of slopes between points

        returns:
            d: list of slopes for the splines
        """
        # Initialize arrays for the coefficients of the tridiagonal matrix
        a = np.zeros(len(h)).astype(float)
        b = np.zeros(len(h)).astype(float)
        c = np.zeros(len(h)).astype(float)
        r = np.zeros(len(h)).astype(float)

        # Set values for the first and second sub-diagonal of the matrix
        a[:-1] = h[1:]  # Set values for all but the last entry of `a`
        a[-1] = h[-2] + h[-1]  # Set value for the last entry of `a`
        b[0] = h[1]  # Set the first value of `b`
        b[1:] = 2 * (h[1:] + h[:-1])  # Set values for all but the first entry of `b`
        b = np.append(b, h[-2])  # Append the value of `h[-2]` to the end of `b`
        c[0] = h[0] + h[1]  # Set the first value of `c`
        c[1:] = h[:-1]  # Set values for all but the first entry of `c`

        # Right-hand side

        # Calculate the first value of the right-hand side
        r[0] = ((h[0] + 2 * c[0]) * h[1] * delta[0] + h[0] ** 2 * delta[1]) / c[0]
        # Calculate values for all but the first entry of the right-hand side
        r[1:] = 3 * (h[1:] * delta[:-1] + h[:-1] * delta[1:])
        # Append a calculated value to the end of the right-hand side
        r = np.append(
            r,
            (h[-1] ** 2 * delta[-2] + (2 * a[-1] + h[-1]) * h[-2] * delta[-1]) / a[-1],
        )

        # Solve the system of equations defined by the tridiagonal matrix and the right-hand side
        res = lu_solve(np.diag(a, -1) + np.diag(b) + np.diag(c, 1), r)

        # Return the solution with type `float`
        return res.astype(float)

    x = np.array(
        x
    )  # Convert the input x data into a numpy array, if it is not already one
    y = np.array(
        y
    )  # Convert the input y data into a numpy array, if it is not already one

    if not sorted:  # If the input points are not sorted in ascending order of x values
        # Sort the points
        ind = np.argsort(x)  # Get the indices of the sorted array after sorting x
        x = x[ind]  # Sort the x coordinates
        y = y[ind]  # Sort the y coordinates

    # First derivative
    h = np.diff(x)  # Calculate the differences between consecutive x values
    delta = (
        np.diff(y) / h
    )  # Calculate the differences between consecutive y values and divide by the differences in x to obtain the first derivative at each point

    d = splineslopes(
        h, delta
    )  # Calculate the slopes of the cubic spline using the helper function `splineslopes` and the calculated differences in x and y

    n = len(x)  # Get the number of points
    c = (3 * delta - 2 * d[:-1] - d[1:]) / (
        h
    )  # Calculate the coefficient c for the cubic spline
    b = (d[:-1] - 2 * delta + d[1:]) / (
        h**2
    )  # Calculate the coefficient b for the cubic spline

    k = np.zeros(
        np.size(u), dtype=int
    )  # Create an array of zeros with the same number of elements as the input u, with data type integer
    for j in np.arange(1, n - 1):  # Loop through all values from 1 to n-1
        k[
            x[j] <= u
        ] = j  # If the j-th value in x is less than or equal to u, set the corresponding value in k to j

    s = (
        u - x[k]
    )  # Calculate the difference between u and the value in x at the corresponding index in k
    v = y[k] + s * (
        d[k] + s * (c[k] + s * b[k])
    )  # Calculate the value of the cubic spline at each point in u using the calculated coefficients and the differences in x and y

    return v  # Return the calculated value of the cubic spline at each point in u



from BNumMet.NonLinear import *
from BNumMet.Visualizers.NonLinearVisualizer import NonLinearVisualizer
import numpy as np
from unittest import TestCase


class test_Roots(TestCase):
    def test_bisect(self):
        """
        Test the bisection method
        """
        f = lambda x: x**2 - 1
        x = bisect(f, [0, 2])
        x1, iters = bisect(f, [0, 2], iters=True)
        self.assertTrue(np.isclose(x, 1))
        self.assertTrue(np.isclose(x1, 1))

    def test_secant(self):
        """
        Test the secant method
        """
        f = lambda x: x**2 - 1
        x = secant(f, [0, 2])
        x1, iters = secant(f, [0, 2], iters=True)
        self.assertTrue(np.isclose(x, 1))
        self.assertTrue(np.isclose(x1, 1))

    def test_newton(self):
        """
        Test the Newton method
        """
        f = lambda x: x**2 - 1
        fprime = lambda x: 2 * x
        x = newton(f, fprime, 3)
        x1, iters = newton(f, fprime, 3, iters=True)
        self.assertTrue(np.isclose(x, 1))
        self.assertTrue(np.isclose(x1, 1))

    def test_zBrentDekker(self):
        """
        Test the zBrentDekker method
        """
        f = lambda x: x**2 - 1
        x0 = zBrentDekker(f, [0, 2])
        x, iters = zBrentDekker(f, [0, 2], iters=True)
        x1, iters1, steps = zBrentDekker(f, [0, 2], iters=True, steps=True)
        x2, steps = zBrentDekker(f, [0, 2], iters=False, steps=True)
        print(x, iters)
        self.assertTrue(np.isclose(x0, 1))
        self.assertTrue(np.isclose(x, 1))
        self.assertTrue(np.isclose(x1, 1))
        self.assertTrue(iters == iters1)
        self.assertTrue(np.isclose(x2, 1))

    def test_IQI(self):
        """
        Test the IQI method
        """
        f = lambda x: x**2 - 1
        x = IQI(f, [0, 2 / 3, 2])
        x1, iters = IQI(f, [0, 2 / 3, 2], iters=True)
        self.assertTrue(np.isclose(x, 1))
        self.assertTrue(np.isclose(x1, 1))

    def test_SameSign(self):
        """
        Test if in (a,b) with both f(a),f(b) >0 or <0 then ValueError is raised in all methods

        """
        f = lambda x: x**2 - 1
        methods = [bisect, secant, zBrentDekker]
        for method in methods:
            with self.assertRaises(ValueError):
                method(f, [-2, 2])

    def test_OneStep_Bisect(self):
        """
        Test if the bisect method returns the correct value after one step, this test is of critical importance, since the future zBrentDekker method is based on the bisect method
        """

        f = lambda x: x**2 - 1
        x = bisect(f, [0, 1.5], 1)
        self.assertTrue(np.isclose(x, 3 / 4))

    def test_OneStep_Secant(self):
        """
        Test if the secant method returns the correct value after one step, this test is of critical importance, since the future zBrentDekker method is based on the secant method
        """
        f = lambda x: x**2 - 1
        x = secant(f, [0, 1.5], 1)
        self.assertTrue(np.isclose(x, 2 / 3))

    def test_OneStep_Newton(self):
        """
        Test if the Newton method returns the correct value after one step
        """
        f = lambda x: x**2 - 1
        fprime = lambda x: 2 * x
        x = newton(f, fprime, 3, 1)
        self.assertTrue(np.isclose(x, 5 / 3))

    def test_OneStep_IQI(self):
        """
        Test if the IQI method returns the correct value after one step, this test is of critical importance, since the future zBrentDekker method is based on the IQI method
        """
        f = lambda x: 6 * x**2 - 2 * x**3
        x = IQI(f, [4, 4.5, 5], 1)
        print(x)
        self.assertTrue(np.isclose(x, 3.3104729014286414))

    def test_zBrentDekker_bookExamples(self):
        """
        Tests the zBrentDekker method on the examples from the reference book
        """
        f1 = (
            lambda x: 0 if abs(x) < 3.8 * 10 ** (-4) else float(x * np.exp(-1 / x**2))
        )
        f2 = (
            lambda x: float(np.exp(x))
            if x > -(10**6)
            else float(np.exp(-(10**6)) - (x + 10**6) ** 2)
        )
        tests = [  # (f, interval, tol, expectedIterations)
            (lambda x: x**9, [-1, 1.1], 10 ** (-9), 81),
            (lambda x: x**9, [-1, 4.0], 10 ** (-20), 189),
            (lambda x: x**19, [-1, 4.0], 10 ** (-20), 195),
            (f1, [-1, 4], 10 ** (-20), 33),
            (f2, [-1001200, 0], 10 ** (-20), 79),
        ]
        for test in tests:
            x, iters = zBrentDekker(test[0], test[1], tol=test[2], iters=True)

            self.assertTrue(np.isclose(test[0](x), 0))
            self.assertLessEqual(iters, test[3])


class Test_NonLinearVisualizer(TestCase):
    def test_no_param_init(self):
        """
        Tests if NonLinearVisualizer is initialized correctly without parameters (i.e. default values are used)
        """

        self.nonLinearVisualizer = NonLinearVisualizer()
        self.nonLinearVisualizer.run()

        fun = lambda x: (x - 1) * (x - 4) * np.exp(-x)
        interval = (0, 3)

        x = np.linspace(interval[0], interval[1], 100)

        self.assertTrue(np.allclose(self.nonLinearVisualizer.f(x), fun(x)))
        self.assertEqual(self.nonLinearVisualizer.a, interval[0])
        self.assertEqual(self.nonLinearVisualizer.b, interval[1])

    def test_acts_as_BrenttDekker(self):
        """
        The NonLinearVisualizer should act as the BrentDekker method if help is followed
        """
        f = lambda x: x**2 - 1
        interval = (0, 2)
        tol = 1e-20
        x, itersbrentt, stack = zBrentDekker(
            f, interval, iters=True, steps=True, tol=tol
        )
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        iters = 0
        guiStack = []
        while nonLinearVisualizer.hint_step != None:
            if nonLinearVisualizer.hint_step == "Bisection":
                guiStack.append("Bisection")
                nonLinearVisualizer.bisect_button.click()
            elif nonLinearVisualizer.hint_step == "Secant":
                guiStack.append("Secant")
                nonLinearVisualizer.secant_button.click()
            elif nonLinearVisualizer.hint_step == "IQI":
                guiStack.append("IQI")
                nonLinearVisualizer.iqi_button.click()
            iters += 1

            if iters > itersbrentt:
                self.fail(
                    "NonLinearVisualizer did not converge to the same value as zBrentDekker"
                )

        self.assertEqual(len(guiStack), len(stack))
        self.assertEqual(guiStack, stack)

    def test_revertStep(self):
        """
        Tests if the revertStep method works correctly
        """
        f = lambda x: x**2 - 1
        interval = (0, 2)
        tol = 1e-20
        x, iters, stack = zBrentDekker(f, interval, iters=True, steps=True, tol=tol)
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        # Perform all steps and store them in a stack
        guiStack = []
        while nonLinearVisualizer.hint_step != None:
            guiStack.append(nonLinearVisualizer.hint_step)
            if nonLinearVisualizer.hint_step == "Bisection":
                nonLinearVisualizer.bisect_button.click()
            elif nonLinearVisualizer.hint_step == "Secant":
                nonLinearVisualizer.secant_button.click()
            elif nonLinearVisualizer.hint_step == "IQI":
                nonLinearVisualizer.iqi_button.click()

        # Revert the last step and check if the hintStep is the same as the future step in the stack
        i = len(guiStack) - 1
        while nonLinearVisualizer.revert_stack != []:
            nonLinearVisualizer.revert_button.click()
            i -= 1
            self.assertTrue(nonLinearVisualizer.hint_step == stack[i + 1])

    def test_reset(self):
        """
        Tests if the reset method works correctly
        """
        f = lambda x: x**2 - 1
        interval = (0, 2)
        tol = 1e-20
        x, iters, stack = zBrentDekker(f, interval, iters=True, steps=True, tol=tol)
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        # Perform all steps and store them in a stack
        guiStack = []
        while nonLinearVisualizer.hint_step != None:
            guiStack.append(nonLinearVisualizer.hint_step)
            if nonLinearVisualizer.hint_step == "Bisection":
                nonLinearVisualizer.bisect_button.click()
            elif nonLinearVisualizer.hint_step == "Secant":
                nonLinearVisualizer.secant_button.click()
            elif nonLinearVisualizer.hint_step == "IQI":
                nonLinearVisualizer.iqi_button.click()

        # A Reset should revert all steps and set the hintStep to the first step in the stack
        nonLinearVisualizer.reset_button.click()
        self.assertEqual(nonLinearVisualizer.revert_stack, [])
        self.assertTrue(nonLinearVisualizer.hint_step == stack[0])

        a = nonLinearVisualizer.a
        c = nonLinearVisualizer.c
        b = nonLinearVisualizer.b

        self.assertEqual(nonLinearVisualizer.a, nonLinearVisualizer.c)
        self.assertListEqual(sorted([a, b]), list(nonLinearVisualizer.original_data))

    def test_legend_length_after_checkbox(self):
        """
        Figure Marks should be more than original after a checkbox is clicked
        """
        f = lambda x: x**2 - 1
        interval = (0, 2)
        tol = 1e-20
        x, iters, stack = zBrentDekker(f, interval, iters=True, steps=True, tol=tol)
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        og_length = len(nonLinearVisualizer.fig.marks)

        nonLinearVisualizer.bisect_checkbox.value = True
        self.assertTrue(len(nonLinearVisualizer.fig.marks) > og_length)

        nonLinearVisualizer.bisect_checkbox.value = False
        self.assertEqual(len(nonLinearVisualizer.fig.marks), og_length)

        nonLinearVisualizer.secant_checkbox.value = True
        self.assertTrue(len(nonLinearVisualizer.fig.marks) > og_length)

        self.assertTrue(
            nonLinearVisualizer.iqi_checkbox.disabled
        )  # IQI is disabled by at first step of this function
        # Perform all steps and store them in a stack
        guiStack = []
        while nonLinearVisualizer.hint_step != None:
            guiStack.append(nonLinearVisualizer.hint_step)
            if nonLinearVisualizer.hint_step == "Bisection":
                nonLinearVisualizer.bisect_button.click()
            elif nonLinearVisualizer.hint_step == "Secant":
                nonLinearVisualizer.secant_button.click()
            elif nonLinearVisualizer.hint_step == "IQI":
                self.assertFalse(nonLinearVisualizer.iqi_checkbox.disabled)
                nonLinearVisualizer.iqi_checkbox.value = True
                self.assertTrue(len(nonLinearVisualizer.fig.marks) > og_length)
                break

    def test_sign_not_guranteed(self):
        """
        Raise exception if signs are the same
        """
        f = lambda x: x**2 - 1
        interval = (2, 4)
        tol = 1e-20
        with self.assertRaises(Exception):
            zBrentDekker(f, interval, tol=tol)

    def test_sign_zero_start(self):
        """
        Already found solution at start
        """
        f = lambda x: x**2 - 1
        interval = (1, 4)
        tol = 1e-20
        x, iters, stack = zBrentDekker(f, interval, iters=True, steps=True, tol=tol)
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        self.assertEqual(x, 1)
        self.assertEqual(iters, 0)
        self.assertEqual(stack, [])

        self.assertEqual(nonLinearVisualizer.b, 1)

    def test_reset_emptyStack(self):
        """
        Test if reset works when the stack is empty (i.e. no steps have been taken)
        """
        f = lambda x: x**2 - 1
        interval = (0, 2)
        tol = 1e-20
        x, iters, stack = zBrentDekker(f, interval, iters=True, steps=True, tol=tol)
        nonLinearVisualizer = NonLinearVisualizer(f, interval, tol=tol)
        nonLinearVisualizer.run()

        # A Reset should revert all steps and set the hintStep to the first step in the stack
        nonLinearVisualizer.reset_button.click()
        self.assertEqual(nonLinearVisualizer.revert_stack, [])
        self.assertTrue(nonLinearVisualizer.hint_step == stack[0])

    def test_exception_wrongSign(self):
        """
        Tests if the exception is raised when the signs are NOT different
        """
        f = lambda x: x**2 - 1
        interval = (2, 4)
        tol = 1e-20
        with self.assertRaises(Exception):
            zBrentDekker(f, interval, tol=tol)
